;; Generated package description from async.el  -*- no-byte-compile: t -*-
(define-package "async" "1.9.3" "Asynchronous processing in Emacs" '((cl-lib "0.5") (nadvice "0.3")) :keywords '("async") :authors '(("John Wiegley" . "jwiegley@gmail.com")) :maintainer '("John Wiegley" . "jwiegley@gmail.com") :url "https://github.com/jwiegley/emacs-async")
